<hr />
<div class="footer">

<a href="{{ welcomeurl }}">Welcome Page</a>
<br />

<a href="http://buildbot.net/">Buildbot</a>-{{ version }}

{% if title %}
  working for
  {% if title_url %}
    <a href="{{ title_url }}">{{ title }}</a>
  {% else %}
    {{ title }}
  {% endif %}
  project
{% endif %}

<br />

Page built: {{ time }} ({{ tz }})
</div>

