<!DOCTYPE html
    PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml"
      xmlns:py="http://genshi.edgewall.org/"
      xmlns:xi="http://www.w3.org/2001/XInclude">
  <xi:include href="layout.html" />
  <xi:include href="macros.html" />
  <head>
    <title py:if="title">$title</title>
    <meta py:if="version" name="ROBOTS" content="NOINDEX, NOFOLLOW" />
  </head>

  <body>
    <div id="content">
      <h1>Active Builders</h1>
      <ol py:if="names">
        <li py:for="name in names">
          <a href="${href.buildbot('builder/'+name)}">${name}</a>
        </li>
      </ol>
    </div>
  </body>
</html>
